"""Device module handling /device/ API calls."""

from rachiopy.rachioobject import RachioObject


class Device(RachioObject):
    """Device class with /device/ API calls."""

    def get(self, dev_id: str):
        """Retrieve the information for a device entity.

        For more info of the content in the response see:
        https://rachio.readme.io/docs/publicdeviceid

        :param dev_id: Device's unique id
        :type dev_id: str

        :return: The return value is a tuple of (response, content), the first
            being and instance of the httplib2.Response class, the second
            being a string that contains the response entity body (Python
            object if it contains JSON).
        :rtype: tuple
        """
        path = f"device/{dev_id}"
        return self.get_request(path)

    def current_schedule(self, dev_id: str):
        """Retrieve current schedule running, if any.

        For more info of the content in the response see:
        https://rachio.readme.io/docs/publicdeviceidcurrent_schedule

        :param dev_id: Device's unique id
        :type dev_id: str

        :return: The return value is a tuple of (response, content), the first
            being and instance of the httplib2.Response class, the second
            being a string that contains the response entity body (Python
            object if it contains JSON).
        :rtype: tuple
        """
        path = f"device/{dev_id}/current_schedule"
        return self.get_request(path)

    def event(self, dev_id: str, starttime: int, endtime: int):
        """Retrieve events for a device entity.

        For more info of the content in the response see:
        https://rachio.readme.io/docs/publicdeviceideventstarttimestarttimeendtimeendtim

        :param dev_id: Device's unique id
        :type dev_id: str

        :param starttime: Query start time milliseconds unix epoch
        :type starttime: int

        :param endtime: Query end time milliseconds unix epoch
        :type endtime: int

        :return: The return value is a tuple of (response, content), the first
            being and instance of the httplib2.Response class, the second
            being a string that contains the response entity body (Python
            object if it contains JSON).
        :rtype: tuple
        """
        path = f"device/{dev_id}/event?startTime={starttime}&endTime={endtime}"
        return self.get_request(path)

    def forecast(self, dev_id: str, units="US"):
        """Retrieve current and predicted forecast.

        For more info of the content in the response see:
        https://rachio.readme.io/docs/publicdeviceidforecastunitsunits

        :param dev_id: Device's unique id
        :type dev_id: str

        :param units: Forecast data units, one of US or METRIC, defaults to US
        :type units: str

        :return: The return value is a tuple of (response, content), the first
            being and instance of the httplib2.Response class, the second
            being a string that contains the response entity body (Python
            object if it contains JSON).
        :rtype: tuple
        """
        assert units in ["US", "METRIC"], "units must be either US or METRIC"
        path = f"device/{dev_id}/forecast?units={units}"
        return self.get_request(path)

    def stop_water(self, dev_id: str):
        """Stop all watering on device.

        For more info of the content in the response see:
        https://rachio.readme.io/docs/devicestop_water

        :param dev_id: Device's unique id
        :type dev_id: str

        :return: The return value is a tuple of (response, content), the first
            being and instance of the httplib2.Response class, the second
            being a string that contains the response entity body (Python
            object if it contains JSON).
        :rtype: tuple
        """
        payload = {"id": dev_id}
        return self.put_request("device/stop_water", payload)

    def rain_delay(self, dev_id: str, duration: int):
        """Rain delay device.

        For more info of the content in the response see:
        https://rachio.readme.io/docs/devicestop_water

        :param dev_id: Device's unique id
        :type dev_id: str

        :param duration: Duration in seconds (Range is 0 - 604800 (7 days) )
        :type duration: int

        :return: The return value is a tuple of (response, content), the first
            being and instance of the httplib2.Response class, the second
            being a string that contains the response entity body (Python
            object if it contains JSON).
        :rtype: tuple
        """
        assert 0 <= duration <= 604800, "duration must be between 0 and 604800"
        payload = {"id": dev_id, "duration": duration}
        return self.put_request("device/rain_delay", payload)

    def turn_on(self, dev_id: str):
        """Turn ON all features of the device.

        schedules, weather intelligence, water budget, etc.

        For more info of the content in the response see:
        https://rachio.readme.io/docs/publicdeviceon-1

        :param dev_id: Device's unique id
        :type dev_id: str

        :return: The return value is a tuple of (response, content), the first
            being and instance of the httplib2.Response class, the second
            being a string that contains the response entity body (Python
            object if it contains JSON).
        :rtype: tuple
        """
        payload = {"id": dev_id}
        return self.put_request("device/on", payload)

    def turn_off(self, dev_id: str):
        """Turn OFF all features of the device.

        schedules, weather intelligence, water budget, etc.

        For more info of the content in the response see:
        https://rachio.readme.io/docs/publicdeviceoff-1

        :param dev_id: Device's unique id
        :type dev_id: str

        :return: The return value is a tuple of (response, content), the first
            being and instance of the httplib2.Response class, the second
            being a string that contains the response entity body (Python
            object if it contains JSON).
        :rtype: tuple
        """
        payload = {"id": dev_id}
        return self.put_request("device/off", payload)

    def pause_zone_run(self, dev_id: str, duration: int):
        """Pause a zone run for device.

        For more info of the content in the response see:
        https://rachio.readme.io/docs/publicdevicepause_zone_run

        :param dev_id: Device's unique id
        :type dev_id: str

        :param duration: Duration in seconds (Range is 0 - 3600 (1 hour) )
        :type duration: int

        :return: The return value is a tuple of (response, content), the first
            being and instance of the httplib2.Response class, the second
            being a string that contains the response entity body (Python
            object if it contains JSON).
        :rtype: tuple
        """
        assert 0 <= duration <= 3600, "duration must be between 0 and 3600"
        payload = {"id": dev_id, "duration": duration}
        return self.put_request("device/pause_zone_run", payload)

    def resume_zone_run(self, dev_id: str):
        """Resume a zone run for device.

        For more info of the content in the response see:
        https://rachio.readme.io/docs/publicdeviceresume_zone_run

        :param dev_id: Device's unique id
        :type dev_id: str

        :return: The return value is a tuple of (response, content), the first
            being and instance of the httplib2.Response class, the second
            being a string that contains the response entity body (Python
            object if it contains JSON).
        :rtype: tuple
        """
        payload = {"id": dev_id}
        return self.put_request("device/resume_zone_run", payload)



"""Person module handling /person/ API calls."""

from rachiopy.rachioobject import RachioObject


class Person(RachioObject):
    """Person class with methods for /person/ API calls."""

    def info(self):
        """Retrieve the id for the person entity currently logged in.

        For more info of the content in the response see:
        https://rachio.readme.io/docs/publicpersoninfo

        :return: The return value is a tuple of (response, content), the first
            being and instance of the httplib2.Response class, the second
            being a string that contains the response entity body (Python
            object if it contains JSON).
        :rtype: tuple
        """
        return self.get_request("person/info")

    def get(self, user_id: str):
        """Retrieve the information for a person entity.

        For more info of the content in the response see:
        https://rachio.readme.io/docs/publicpersonid

        :param user_id: Person's unique id
        :type user_id: str

        :return: The return value is a tuple of (response, content), the first
            being and instance of the httplib2.Response class, the second
            being a string that contains the response entity body (Python
            object if it contains JSON).
        :rtype: tuple
        """
        path = f"person/{user_id}"
        return self.get_request(path)



"""Flexschedulerule module handling /flexschedulerule/ API calls."""

from rachiopy.rachioobject import RachioObject


class FlexSchedulerule(RachioObject):
    """FlexSchedulerule class with methods for /flexschedulerule/ calls."""

    def get(self, flex_sched_rule_id: str):
        """Retrieve the information for a flexscheduleRule entity.

        For more info of the content in the response see:
        https://rachio.readme.io/docs/publicflexscheduleruleid

        :param flex_sched_rule_id: FlexScheduleRule's unique id
        :type flex_sched_rule_id: str

        :return: The return value is a tuple of (response, content), the first
            being and instance of the httplib2.Response class, the second
            being a string that contains the response entity body.
        :rtype: tuple
        """
        path = f"flexschedulerule/{flex_sched_rule_id}"
        return self.get_request(path)



"""RachioObject module containing a helper class for all API calls."""

from requests import Session

_API_URL = "https://api.rach.io/1/public"


class RachioObject:
    """The Rachio base object."""

    def __init__(self, authtoken: str, http_session=None):
        """Rachioobject class initializer.

        :param authtoken: The API authentication token.
        :type authtoken: str

        :param http_session: The HTTP Session
        :type http_session: Session
        """
        self.authtoken = authtoken
        self._headers = {
            "Content-Type": "application/json",
            "Authorization": f"Bearer {authtoken}",
        }
        self._http_session = http_session or Session()

    def _request(self, path: str, method: str, body=None):
        """Make a request from the API.

        :return: The return value is a tuple of (response, content), the first
            being and instance of the httplib2.Response class, the second
            being a string that contains the response entity body (Python
            object if it contains JSON).
        :rtype: tuple
        """
        url = f"{_API_URL}/{path}"
        response = self._http_session.request(
            method, url, headers=self._headers, data=body
        )

        content_type = response.headers.get("content-type")
        headers = {k.lower(): v for k, v in response.headers.items()}
        headers["status"] = response.status_code

        if content_type and content_type.startswith("application/json"):
            return headers, response.json()

        return headers, response.text

    def get_request(self, path: str, body=None):
        """Make a GET request to the API.

        :return: The return value is a tuple of (response, content), the first
            being and instance of the httplib2.Response class, the second
            being a string that contains the response entity body (Python
            object if it contains JSON).
        :rtype: tuple
        """
        return self._request(path, "GET", body)

    def put_request(self, path: str, body=None):
        """Make a PUT request to the API.

        :return: The return value is a tuple of (response, content), the first
            being and instance of the httplib2.Response class, the second
            being a string that contains the response entity body (Python
            object if it contains JSON).
        :rtype: tuple
        """
        return self._request(path, "PUT", body)

    def post_request(self, path: str, body=None):
        """Make a POST request to the API.

        :return: The return value is a tuple of (response, content), the first
            being and instance of the httplib2.Response class, the second
            being a string that contains the response entity body (Python
            object if it contains JSON).
        :rtype: tuple
        """
        return self._request(path, "POST", body)

    def delete_request(self, path: str, body=None):
        """Make a DELETE request to the API.

        :return: The return value is a tuple of (response, content), the first
            being and instance of the httplib2.Response class, the second
            being a string that contains the response entity body (Python
            object if it contains JSON).
        :rtype: tuple
        """
        return self._request(path, "DELETE", body)



"""Notification module handling /notification/ API calls."""

from rachiopy.rachioobject import RachioObject


class Notification(RachioObject):
    """Notification class with methods for /notification/ API calls."""

    def get_webhook_event_type(self):
        """Retrieve the list of events types.

        Event types that are available to any webhook for subscription.

        For more info of the content in the response see:
        https://rachio.readme.io/docs/publicnotificationwebhook_event_type

        :return: The return value is a tuple of (response, content), the first
            being and instance of the httplib2.Response class, the second
            being a string that contains the response entity body.
        :rtype: tuple
        """
        return self.get_request("notification/webhook_event_type")

    def get_device_webhook(self, dev_id: str):
        """Retrieve all webhooks for a device.

        For more info of the content in the response see:
        https://rachio.readme.io/docs/publicnotificationdeviceidwebhook

        :param dev_id: Device's unique id
        :type dev_id: str

        :return: The return value is a tuple of (response, content), the first
            being and instance of the httplib2.Response class, the second
            being a string that contains the response entity body.
        :rtype: tuple
        """
        path = f"notification/{dev_id}/webhook"
        return self.get_request(path)

    def add(self, dev_id: str, external_id: str, url: str, event_types):
        """Add a webhook to a device.

        externalId can be used as opaque data that
        is tied to your company, and passed back in each webhook event
        response.

        For more info of the content in the response see:
        https://rachio.readme.io/docs/publicnotificationwebhook

        :param dev_id: Device's unique id
        :type dev_id: str

        :param external_id: External company ID
        :type exteranl_id: str

        :param url: External webhook URL
        :type url: str

        :param event_type: Event types
        :type event_type: Object[]

        :return: The return value is a tuple of (response, content), the first
            being and instance of the httplib2.Response class, the second
            being a string that contains the response entity body.
        :rtype: tuple
        """
        payload = {
            "device": {"id": dev_id},
            "externalId": external_id,
            "url": url,
            "eventTypes": event_types,
        }
        return self.post_request("notification/webhook", payload)

    def update(self, hook_id: str, external_id: str, url: str, event_types):
        """Update a webhook.

        For more info of the content in the response see:
        https://rachio.readme.io/docs/publicnotificationwebhook-1

        :param hook_id: Webhook's unique id
        :type hook_id: str

        :param external_id: External company ID
        :type exteranl_id: str

        :param url: External webhook URL
        :type url: str

        :param event_type: Event types
        :type event_type: Object[]

        :return: The return value is a tuple of (response, content), the first
            being and instance of the httplib2.Response class, the second
            being a string that contains the response entity body.
        :rtype: tuple
        """
        payload = {
            "id": hook_id,
            "externalId": external_id,
            "url": url,
            "eventTypes": event_types,
        }
        return self.put_request("notification/webhook", payload)

    def delete(self, hook_id: str):
        """Remove a webhook.

        For more info of the content in the response see:
        https://rachio.readme.io/docs/publicnotificationwebhookid

        :param hook_id: Webhook's unique id
        :type hook_id: str

        :return: The return value is a tuple of (response, content), the first
            being and instance of the httplib2.Response class, the second
            being a string that contains the response entity body.
        :rtype: tuple
        """
        path = f"notification/webhook/{hook_id}"
        return self.delete_request(path)

    def get(self, hook_id: str):
        """Get a webhook.

        For more info of the content in the response see:
        https://rachio.readme.io/docs/publicnotificationdeviceidwebhook

        :param hook_id: Webhook's unique id
        :type hook_id: str

        :return: The return value is a tuple of (response, content), the first
            being and instance of the httplib2.Response class, the second
            being a string that contains the response entity body.
        :rtype: tuple
        """
        path = f"notification/webhook/{hook_id}"
        return self.get_request(path)



"""Zone module handling /zone/ API calls."""

from decimal import Decimal

from rachiopy.rachioobject import RachioObject


class Zone(RachioObject):
    """Zone class with methods for /zone/ API calls."""

    def start(self, zone_id: str, duration: int):
        """Start a zone.

        For more info of the content in the response see:
        https://rachio.readme.io/docs/zonestart

        :param zone_id: Zone's unique id
        :type zone_id: str

        :param duration: Duration in seconds (Range is 0 - 10800 (3 Hours) )
        :type duration: int

        :return: The return value is a tuple of (response, content), the first
            being and instance of the httplib2.Response class, the second
            being a string that contains the response entity body (Python
            object if it contains JSON).
        :rtype: tuple
        """
        assert 0 <= duration <= 10800, "duration must be in range 0-10800"
        payload = {"id": zone_id, "duration": duration}
        return self.put_request("zone/start", payload)

    def start_multiple(self, zones):
        """Start multiple zones.

        For more info of the content in the response see:
        https://rachio.readme.io/docs/publiczonestart_multiple

        :param zones: Zone's unique id, duration, and sort order
        :type zones: Object[]

        :return: The return value is a tuple of (response, content), the first
            being and instance of the httplib2.Response class, the second
            being a string that contains the response entity body (Python
            object if it contains JSON).
        :rtype: tuple
        """
        payload = {"zones": zones}
        return self.put_request("zone/start_multiple", payload)

    def schedule(self):
        """Create an empty zone schedule."""
        return ZoneSchedule(self)

    def set_moisture_percent(self, zone_id: str, percent: Decimal):
        """Set zone moisture percent.

        For more info of the content in the response see:
        https://rachio.readme.io/docs/publiczonesetmoisturepercent

        :param zone_id: Zone's unique id
        :type zone_id: str

        :param percent: Soil moisture percent (Range is 0 - 1 )
        :type percent: Decimal

        :return: The return value is a tuple of (response, content), the first
            being and instance of the httplib2.Response class, the second
            being a string that contains the response entity body (Python
            object if it contains JSON).
        :rtype: tuple
        """
        assert 0 <= percent <= 1, "percent must be in range 0.0-1.0"
        payload = {"id": zone_id, "percent": percent}
        return self.put_request("zone/setMoisturePercent", payload)

    def set_moisture_level(self, zone_id: str, level: Decimal):
        """Set zone moisture level.

        For more info of the content in the response see:
        https://rachio.readme.io/docs/publiczonesetmoisturelevel

        :param zone_id: Zone's unique id
        :type zone_id: str

        :param level: Soil moisture level in mm (Range is 0 - Maximum Moisture
        in mm (depth of water + (10% depth of water))
        :type level: Decimal

        :return: The return value is a tuple of (response, content), the first
            being and instance of the httplib2.Response class, the second
            being a string that contains the response entity body (Python
            object if it contains JSON).
        :rtype: tuple
        """
        payload = {"id": zone_id, "level": level}
        return self.put_request("zone/setMoistureLevel", payload)

    def get(self, zone_id: str):
        """Retrieve the information for a zone entity.

        For more info of the content in the response see:
        https://rachio.readme.io/docs/publiczoneid

        :param zone_id: Zone's unique id
        :type zone_id: str

        :return: The return value is a tuple of (response, content), the first
            being and instance of the httplib2.Response class, the second
            being a string that contains the response entity body (Python
            object if it contains JSON).
        :rtype: tuple
        """
        path = f"zone/{zone_id}"
        return self.get_request(path)


class ZoneSchedule:
    """Help with starting multiple zones."""

    def __init__(self, zone_api: Zone):
        """Zoneschedule class initializer."""
        self._api = zone_api
        self._zones = []

    def enqueue(self, zone_id: str, duration: int):
        """Add a zone and duration to the schedule.

        :param zone_id: Zone's unique id
        :type zone_id: str

        :param duration: Duration in seconds (Range is 0 - 10800 (3 Hours) )
        :type duration: int
        """
        self._zones.append((zone_id, duration))

    def start(self):
        """Start the schedule."""
        zones = [
            {"id": data[0], "duration": data[1], "sortOrder": count}
            for (count, data) in enumerate(self._zones, 1)
        ]
        self._api.start_multiple(zones)

    def __enter__(self):
        """Allow a schedule to be created in a with block."""
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Allow the schedule to be executed by leaving with block."""
        self.start()



"""Main rachiopy module."""

from rachiopy.rachioobject import RachioObject
from rachiopy.person import Person
from rachiopy.device import Device
from rachiopy.flexschedulerule import FlexSchedulerule
from rachiopy.notification import Notification
from rachiopy.schedulerule import Schedulerule
from rachiopy.zone import Zone


class Rachio(RachioObject):
    """Object representing the Rachio API."""

    def __init__(self, authtoken: str):
        """Initialze the Rachio API wrapper.

        :param authtoken: The API authentication token.
        :type authtoken: str
        """
        super().__init__(authtoken)
        self.person = Person(authtoken)
        self.device = Device(authtoken)
        self.flexschedulerule = FlexSchedulerule(authtoken)
        self.notification = Notification(authtoken)
        self.schedulerule = Schedulerule(authtoken)
        self.zone = Zone(authtoken)



"""Schedulerule module handling /scheduerule/ API calls."""

from decimal import Decimal

from rachiopy.rachioobject import RachioObject


class Schedulerule(RachioObject):
    """Schedulerule class with methods for /schedulerule/ API calls."""

    def skip(self, sched_rule_id: str):
        """Skip a schedule rule (watering time).

        For more info of the content in the response see:
        https://rachio.readme.io/docs/scheduleruleskip

        :param sched_rule_id: Schedule rule's unique id
        :type sched_rule_id: str

        :return: The return value is a tuple of (response, content), the first
            being and instance of the httplib2.Response class, the second
            being a string that contains the response entity body.
        :rtype: tuple
        """
        payload = {"id": sched_rule_id}
        return self.put_request("schedulerule/skip", payload)

    def start(self, sched_rule_id: str):
        """Start a schedule rule (watering time).

        For more info of the content in the response see:
        https://rachio.readme.io/docs/schedulerulestart

        :param sched_rule_id: Schedule rule's unique id
        :type sched_rule_id: str

        :return: The return value is a tuple of (response, content), the first
            being and instance of the httplib2.Response class, the second
            being a string that contains the response entity body.
        :rtype: tuple
        """
        payload = {"id": sched_rule_id}
        return self.put_request("schedulerule/start", payload)

    def seasonal_adjustment(self, sched_rule_id: str, adjustment: Decimal):
        """Seasonal adjustment for a schedule rule (watering time).

        This adjustment amount will be applied to the overall run time of the
        selected schedule while overriding any current adjustments.

        For more info of the content in the response see:
        https://rachio.readme.io/docs/publicscheduleruleseasonal_adjustment

        :param sched_rule_id: Schedule rule's unique id
        :type sched_rule_id: str

        :param adjustment: Seasonal adjustment percent from 100% to -100%
            (valid data range is 1 to -1)
        :type adjustment: Decimal

        :return: The return value is a tuple of (response, content), the first
            being and instance of the httplib2.Response class, the second
            being a string that contains the response entity body.
        :rtype: tuple
        """
        payload = {"id": sched_rule_id, "adjustment": adjustment}
        return self.put_request("schedulerule/seasonal_adjustment", payload)

    def get(self, sched_rule_id: str):
        """Retrieve the information for a scheduleRule entity.

        For more info of the content in the response see:
        https://rachio.readme.io/docs/publicscheduleruleid

        :param sched_rule_id: Schedule rule's unique id
        :type sched_rule_id: str

        :return: The return value is a tuple of (response, content), the first
            being and instance of the httplib2.Response class, the second
            being a string that contains the response entity body.
        :rtype: tuple
        """
        path = f"schedulerule/{sched_rule_id}"
        return self.get_request(path)



"""Rachiopy setup script."""
from setuptools import find_packages, setup

VERSION = "1.0.0"

GITHUB_USERNAME = "rfverbruggen"
GITHUB_REPOSITORY = "rachiopy"

GITHUB_PATH = f"{GITHUB_USERNAME}/{GITHUB_REPOSITORY}"
GITHUB_URL = f"https://github.com/{GITHUB_PATH}"

DOWNLOAD_URL = f"{GITHUB_URL}/archive/{VERSION}.tar.gz"
PROJECT_URLS = {"Bug Reports": f"{GITHUB_URL}/issues"}

PACKAGES = find_packages(exclude=["tests", "tests.*"])

setup(
    name="RachioPy",
    version=VERSION,
    author="Robbert Verbruggen",
    author_email="rfverbruggen@icloud.com",
    packages=PACKAGES,
    install_requires=["requests"],
    url=GITHUB_URL,
    download_url=DOWNLOAD_URL,
    project_urls=PROJECT_URLS,
    license="MIT",
    description="A Python module for the Rachio API.",
    platforms="Cross Platform",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Topic :: Software Development",
    ],
)
